<?php

require_once 'SoapHelper.php';
require_once 'AUSKey.php';

class VanguardClient extends SoapHelper {    
    const ACTION_ISSUE = 'http://docs.oasis-open.org/ws-sx/ws-trust/200512/RST/Issue';

    private $auskey;
	protected $wsdl = VANURL;
    
    public function __construct() {
        parent::__construct();
    }
    
    public function requestToken($url) {	
		if (empty($this->auskey)) {
			throw new Exception("AUSKey is empty.");
		}
		
		try {
			$xml = $this->buildRSTdocument($url);
		} catch (Exception $ex) {
			throw $ex;
		}
		try {
			$response = $this->sendSoapRequest(self::ACTION_ISSUE, $xml);
		} catch (Exception $ex) {
			throw $ex;
		}
		
		try {
			$token = $this->parseTokenResponse($response);
		} catch (Exception $ex) {
			throw $ex;
		}
		
		return $token;
	}
    	
	private function buildRSTdocument($url) {
        $envelope = $this->doc->createElementNS(self::NS_SOAP, self::NS_SOAP_PREFIX.':Envelope');

        $this->doc->appendChild($envelope);
        
        try {
            $header = $this->buildSoapHeader();
        } catch (Exception $ex) {
            throw $ex;
        }
        
        $envelope->appendChild($header);
        $envelope->appendChild($this->buildSoapBody($url));
        
        $xml = $this->doc->saveXML();
        return $this->canonicalize($xml);
	}
    
    private function buildSoapHeader() {
        $header = $this->doc->createElementNS(self::NS_SOAP, self::NS_SOAP_PREFIX.':Header');
        
        $action = $this->doc->createElementNS(self::NS_ADDR, self::NS_ADDR_PREFIX.':Action', self::ACTION_ISSUE);
        $action->setAttributeNS(self::NS_SOAP, self::NS_SOAP_PREFIX.':mustUnderstand', 1);
        $header->appendChild($action);
        
        $header->appendChild($this->doc->createElementNS(self::NS_ADDR, self::NS_ADDR_PREFIX.':MessageID', $this->message_id));
        
        $replyto = $this->doc->createElementNS(self::NS_ADDR, self::NS_ADDR_PREFIX.':ReplyTo');
        $replyto->appendChild($this->doc->createElementNS(self::NS_ADDR, self::NS_ADDR_PREFIX.':Address', 'http://www.w3.org/2005/08/addressing/anonymous'));
        $header->appendChild($replyto);
        
        $header->appendChild($this->buildTo());
        
        try {
            $sh = $this->buildSecurityHeader();
        } catch (Exception $ex) {
            throw $ex;
        }
        $header->appendChild($sh);
        
        return $header;
    }
	
	private function buildSecurityHeader() {
        $s = $this->doc->createElementNS(self::NS_SEC, self::NS_SEC_PREFIX.':Security');
        $s->setAttributeNS(self::NS_SOAP, self::NS_SOAP_PREFIX.':mustUnderstand', 1);
        
        $s->appendChild($this->buildTimestampHeader());
        
        $bst = $this->doc->createElementNS(self::NS_SEC, self::NS_SEC_PREFIX.':BinarySecurityToken', $this->getBinarySecurityToken());
        $bst->setAttributeNS(self::NS_WSU, self::NS_WSU_PREFIX.':Id', $this->token_id);
        $bst->setAttribute('ValueType', 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-x509-token-profile-1.0#X509v3');
        $bst->setAttribute('EncodingType', 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-soap-message-security-1.0#Base64Binary');
        $s->appendChild($bst);
        
        try {
            $signature = $this->buildSignature();
        } catch (Exception $ex) {
            throw $ex;
        }
        
        $s->appendChild($signature);
        
        return $s;
	}
    
    private function buildSoapBody($url) { 
	    global $lms;
	    global $database;
	    $ssid = array_sum(str_split($lms->current_user->entity->id)) % 10;
		$ssid = str_pad($lms->current_user->entity->id . $ssid, 10, '0', STR_PAD_LEFT);
		$abn = str_replace(' ', '', $lms->current_user->entity->abn);
                       
        $body = $this->doc->createElementNS(self::NS_SOAP, self::NS_SOAP_PREFIX.':Body');
        
        $rst = $this->doc->createElementNS(self::NS_WST, self::NS_WST_PREFIX.':RequestSecurityToken');
        $rst->appendChild($this->doc->createElementNS(self::NS_WST, self::NS_WST_PREFIX.':RequestType', 'http://docs.oasis-open.org/ws-sx/ws-trust/200512/Issue'));
        
        $rst->appendChild($this->doc->createElementNS(self::NS_WST, self::NS_WST_PREFIX.':KeyType', 'http://docs.oasis-open.org/ws-sx/ws-trust/200512/SymmetricKey'));
        $rst->appendChild($this->doc->createElementNS(self::NS_WST, self::NS_WST_PREFIX.':KeySize', '256'));
        
        $clms = $this->doc->createElementNS(self::NS_WST, self::NS_WST_PREFIX.':Claims');
        $clms->setAttribute('Dialect', 'http://schemas.xmlsoap.org/ws/2005/05/identity');
        $clms->setAttributeNS('http://www.w3.org/2000/xmlns/' ,'xmlns:i', 'http://schemas.xmlsoap.org/ws/2005/05/identity');

        
        $clm = $this->doc->createElementNS('http://schemas.xmlsoap.org/ws/2005/05/identity', 'i:ClaimType');
        $clm->setAttribute('Uri', 'http://vanguard.ebusiness.gov.au/2008/06/identity/claims/abn');
        //$clm->setAttribute('Optional', 'false');
        $clms->appendChild($clm);
        
        $clm = $this->doc->createElementNS('http://schemas.xmlsoap.org/ws/2005/05/identity', 'i:ClaimType');
        $clm->setAttribute('Uri', 'http://vanguard.ebusiness.gov.au/2008/06/identity/claims/credentialtype');
        //$clm->setAttribute('Optional', 'false');
        $clms->appendChild($clm);
        
        $clm = $this->doc->createElementNS('http://schemas.xmlsoap.org/ws/2005/05/identity', 'i:ClaimType');
        $clm->setAttribute('Uri', 'http://vanguard.ebusiness.gov.au/2008/06/identity/claims/samlsubjectid');
        //$clm->setAttribute('Optional', 'false');
        $clms->appendChild($clm);
        
        $rst->appendChild($clms);
        
        $rst->appendChild($this->doc->createElementNS(self::NS_WST, self::NS_WST_PREFIX.':KeyWrapAlgorithm', 'http://www.w3.org/2001/04/xmlenc#rsa-oaep-mgf1p'));
        $rst->appendChild($this->doc->createElementNS(self::NS_WST, self::NS_WST_PREFIX.':EncryptWith', 'http://www.w3.org/2001/04/xmlenc#aes256-cbc'));
        $rst->appendChild($this->doc->createElementNS(self::NS_WST, self::NS_WST_PREFIX.':SignWith', 'http://www.w3.org/2000/09/xmldsig#hmac-sha1'));
        $rst->appendChild($this->doc->createElementNS(self::NS_WST, self::NS_WST_PREFIX.':CanonicalizationAlgorithm', 'http://www.w3.org/2001/10/xml-exc-c14n#'));
        $rst->appendChild($this->doc->createElementNS(self::NS_WST, self::NS_WST_PREFIX.':EncryptionAlgorithm', 'http://www.w3.org/2001/04/xmlenc#aes256-cbc'));
        
        $appliesto = $this->doc->createElementNS(self::NS_WSP, self::NS_WSP_PREFIX.':AppliesTo');
        $epr = $this->doc->createElementNS(self::NS_ADDR, self::NS_ADDR_PREFIX.':EndpointReference');
        $epr->appendChild($this->doc->createElementNS(self::NS_ADDR, self::NS_ADDR_PREFIX.':Address', $url));
        $appliesto->appendChild($epr);
        
        $rst->appendChild($appliesto);
                
        $act = $this->doc->createElementNS(self::NS_TR, self::NS_TR_PREFIX.':ActAs');
        
        $rlt = $this->doc->createElementNS(self::NS_V13, self::NS_V13_PREFIX.':RelationshipToken');
        $rlt->setAttribute('ID', '100001');
        
        $rel = $this->doc->createElementNS(self::NS_V13, self::NS_V13_PREFIX.':Relationship');
        $rel->setAttributeNS(self::NS_V13, self::NS_V13_PREFIX.':Type', 'OSPfor');
        
        $att = $this->doc->createElementNS(self::NS_V13, self::NS_V13_PREFIX.':Attribute');
        $att->setAttributeNS(self::NS_V13, self::NS_V13_PREFIX.':Name', 'SSID');
        $att->setAttributeNS(self::NS_V13, self::NS_V13_PREFIX.':Value', SSID); 
        
        $rel->appendChild($att);
        $rlt->appendChild($rel);
        
        $fp = $this->doc->createElementNS(self::NS_V13, self::NS_V13_PREFIX.':FirstParty');
        $fp->setAttributeNS(self::NS_V13, self::NS_V13_PREFIX.':Scheme', 'uri://abr.gov.au/ABN');
        $fp->setAttributeNS(self::NS_V13, self::NS_V13_PREFIX.':Value', FIRST);
        $rlt->appendChild($fp);
        
        $sp = $this->doc->createElementNS(self::NS_V13, self::NS_V13_PREFIX.':SecondParty');
        $sp->setAttributeNS(self::NS_V13, self::NS_V13_PREFIX.':Scheme', 'uri://abr.gov.au/ABN');
        $sp->setAttributeNS(self::NS_V13, self::NS_V13_PREFIX.':Value', SECOND);
        $rlt->appendChild($sp);
        
        $act->appendChild($rlt);
        $rst->appendChild($act);
        
        $rst->appendChild($this->doc->createElementNS(self::NS_WST, self::NS_WST_PREFIX.':ComputedKeyAlgorithm', 'http://docs.oasis-open.org/ws-sx/ws-trust/200512/CK/PSHA1'));
        
        $body->appendChild($rst);
        
        return $body;
    }
	
	public function loadAuskey($path, $password = '') {
		if (!file_exists($path)) {
			throw new Exception("AUSKey not found at the path specified in Settings.");
		}

		$xml = file_get_contents($path);
		try {
			$this->setAuskey($xml, $password);
		} catch (Exception $ex) {
			throw $ex;
		}
	}
	
	public function setAuskey($xml, $password = '') {
		try {
			$this->auskey = new AUSKey($xml, $password);
		} catch (Exception $ex) {
			throw $ex;
		}
		
		$this->auskey->setCredential(0);
	}
	
	/**
	 * @returns AUSKey
	 */
	public function getAuskey() {
		return $this->auskey;
	}

	private function buildSignature() {
        $signedinfo = $this->buildSignedInfo();
        
		try {
            $signature = $this->getSignatureHash($signedinfo);
		} catch (Exception $ex) {
			throw $ex;
		}
        
        $s = $this->doc->createElementNS(self::NS_SIGN, self::NS_SIGN_PREFIX.':Signature');
        
        $s->appendChild($signedinfo);
        
        $sv = $this->doc->createElementNS(self::NS_SIGN, self::NS_SIGN_PREFIX.':SignatureValue', $signature);
        $s->appendChild($sv);
        
        $keyinfo = $this->doc->createElementNS(self::NS_SIGN, self::NS_SIGN_PREFIX.':KeyInfo');
        $str = $this->doc->createElementNS(self::NS_SEC, self::NS_SEC_PREFIX.':SecurityTokenReference');
        $ref = $this->doc->createElementNS(self::NS_SEC, self::NS_SEC_PREFIX.':Reference');
        $ref->setAttribute('ValueType', 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-x509-token-profile-1.0#X509v3');
        $ref->setAttribute('URI', '#'.$this->token_id);
        $str->appendChild($ref);
        $keyinfo->appendChild($str);
        $s->appendChild($keyinfo);
        
        return $s;
	}
    
    private function buildSignedInfo()
    { 
        $this->buf1 = $this->doc->saveXML($this->buildTimestampHeader());
        $this->cbuf1 = $this->canonicalize($this->buf1);
        $this->hbuf1 = hash('sha256', $this->cbuf1, true);
        $this->digest1 = base64_encode($this->hbuf1);
        
        $this->buf2 = $this->doc->saveXML($this->buildTo());
        $this->cbuf2 = $this->canonicalize($this->buf2);
        $this->hbuf2 = hash('sha256', $this->cbuf2, true);
        $this->digest2 = base64_encode($this->hbuf2);
        
        $buf = $this->doc->saveXML($this->buildTimestampHeader());
		$digest1 = base64_encode(hash('sha256', $this->canonicalize($buf), true));
		
		$buf = $this->doc->saveXML($this->buildTo());
		$digest2 = base64_encode(hash('sha256', $this->canonicalize($buf), true));
        
        $si = $this->doc->createElementNS(self::NS_SIGN, self::NS_SIGN_PREFIX.':SignedInfo');
        
        $cm = $this->doc->createElementNS(self::NS_SIGN, self::NS_SIGN_PREFIX.':CanonicalizationMethod');
        $cm->setAttribute('Algorithm', 'http://www.w3.org/2001/10/xml-exc-c14n#');
        $si->appendChild($cm);
        
        $sm = $this->doc->createElementNS(self::NS_SIGN, self::NS_SIGN_PREFIX.':SignatureMethod');
        $sm->setAttribute('Algorithm', 'http://www.w3.org/2001/04/xmldsig-more#rsa-sha256');
        $si->appendChild($sm);
        
        $transforms = $this->doc->createElementNS(self::NS_SIGN, self::NS_SIGN_PREFIX.':Transforms');
        $transform = $this->doc->createElementNS(self::NS_SIGN, self::NS_SIGN_PREFIX.':Transform');
        $transform->setAttribute('Algorithm', 'http://www.w3.org/2001/10/xml-exc-c14n#');
        $transforms->appendChild($transform);
        $digestmethod = $this->doc->createElementNS(self::NS_SIGN, self::NS_SIGN_PREFIX.':DigestMethod');
        $digestmethod->setAttribute('Algorithm', 'http://www.w3.org/2001/04/xmlenc#sha256');
        
        $digestvalue = $this->doc->createElementNS(self::NS_SIGN, self::NS_SIGN_PREFIX.':DigestValue', $digest1);
        $ref1 = $this->doc->createElementNS(self::NS_SIGN, self::NS_SIGN_PREFIX.':Reference');
        $ref1->setAttribute('URI', '#_0');
        $ref1->appendChild($transforms);
        $ref1->appendChild($digestmethod);
        $ref1->appendChild($digestvalue);
        $si->appendChild($ref1);
        
        $transforms = $this->doc->createElementNS(self::NS_SIGN, self::NS_SIGN_PREFIX.':Transforms');
        $transform = $this->doc->createElementNS(self::NS_SIGN, self::NS_SIGN_PREFIX.':Transform');
        $transform->setAttribute('Algorithm', 'http://www.w3.org/2001/10/xml-exc-c14n#');
        $transforms->appendChild($transform);
        $digestmethod = $this->doc->createElementNS(self::NS_SIGN, self::NS_SIGN_PREFIX.':DigestMethod');
        $digestmethod->setAttribute('Algorithm', 'http://www.w3.org/2001/04/xmlenc#sha256');
        
        $digestvalue = $this->doc->createElementNS(self::NS_SIGN, self::NS_SIGN_PREFIX.':DigestValue', $digest2);
        $ref2 = $this->doc->createElementNS(self::NS_SIGN, self::NS_SIGN_PREFIX.':Reference');
        $ref2->setAttribute('URI', '#_1');
        $ref2->appendChild($transforms);
        $ref2->appendChild($digestmethod);
        $ref2->appendChild($digestvalue);
        $si->appendChild($ref2);
        
        return $si;
    }
	
	
	
	private function parseTokenResponse(&$response) {
		try {
			$xml = new SimpleXMLElement($response);
		} catch (Exception $ex) {
			throw $ex;
		}
		
		$ns = $xml->getNamespaces(true);
		$soap = $xml->children($ns['soap']);
		
		if (!empty($soap->Body->Fault)) {
            $fault = $soap->Body->Fault->children($ns['soap']);
			throw new Exception(reset($fault->Reason->Text));
		}
		
		return $response;
	}
	
    private function getBinarySecurityToken() {
        /* An insanely barbaric way to get a public cert out of PKCS7 chain 
         * which we have inside an AUSkey file. 
         * Apparently, you can't do it any other way with native PHP functions!
         */
        
        // First, we have to dump PKCS7 "publicCertificate" part into a temp file.
        $pkcs7_file = '/tmp/pkcs7';
        $f = fopen($pkcs7_file, 'w');
        fwrite($f, $this->getPublicCertificatePEM());
        fclose($f);
        
        // Then we extract the certificates using 'openssl' utility.
        $keys = shell_exec('openssl pkcs7 -print_certs -in '. $pkcs7_file);
        unlink($pkcs7_file); // We don't want this hanging around in the file system!!!
        if (empty($keys)) {
            return null;
        }
        
        // Then we parse the response (should be 3 PEM certs there).
        $matches = null;
        preg_match_all('/-----BEGIN CERTIFICATE-----(.*)-----END CERTIFICATE-----/mUs', $keys, $matches);
        if (empty($matches)) {
            return null;
        }
        
        /* We have 3 certs here:
         * 
         * 1) Signature Algorithm: sha1WithRSAEncryption
         * 2) Signature Algorithm: sha256WithRSAEncryption
         * 3) Signature Algorithm: sha512WithRSAEncryption
         * 
         * We need the first one (index 0 in $matches[1]).
         * Clean it up and return.
         */
        
        $search = array("\n", "\r",);
        $bst = str_replace($search, '', $matches[1][0]);
        
        return $bst;
    }
    
    public function getPublicCertificatePEM() {
		$key  = "-----BEGIN PKCS7-----\n";
		$key .= chunk_split($this->getAuskey()->getCredential()->publicCertificate, 64, "\n");
		$key .= "-----END PKCS7-----";
        
		return $key;		
    }
	
	public function getPrivateKeyPEM() {
		$key  = "-----BEGIN ENCRYPTED PRIVATE KEY-----\n";
		$key .= chunk_split($this->getAuskey()->getCredential()->protectedPrivateKey, 64, "\n");
		$key .= "-----END ENCRYPTED PRIVATE KEY-----";
        
		return $key;		
	}
	
	private function getSignatureHash($signedinfo) {
        
		$pk = openssl_pkey_get_private($this->getPrivateKeyPEM(), $this->getAuskey()->getPassword());
        if (!$pk) {
            throw new Exception('Unable to access Private Key.');
        }
        
		$xml = $this->canonicalize($this->doc->saveXML($signedinfo));
        $signature = ''; // to be filled below
        
		$signed = openssl_sign($xml, $signature, $pk, OPENSSL_ALGO_SHA256);
        if (!$signed) {
            throw new Exception('Unable to sign the request using this Private Key.');
        }
		openssl_free_key($pk);
        
		return base64_encode($signature);
	}

}


